//*******************************************************************
//
//  UNIT:     D a n c e                           03.09.99,  09.09.99
//
//  dancing RGB lights
//
//-------------------------------------------------------------------
//
//  StartLissaDance        start lissa dance demo
//  ResizeLightningMap     called on new window size
//  LissaDance             animated RGB light demo
//
//*******************************************************************
unit Dance;

interface

function StartLissaDance: string;
procedure ResizeLightningMap;
procedure LissaDance;

function LPP (y:integer): pointer;
//*********************************************************
implementation

uses  SysUtils, Windows, Graphics,
      GrForm, Graph, Palette;

var   lightBm: tBitmap = nil;

const minLI =  33;      // minimum of light intensity
      maxLI = 200;      // maximum of light intensity

var   moveSpeed: integer = 30;    // light moving speed
      radius:    integer = 65;    // moving radius

//*********************************************************
function LPP (y:integer): pointer;
begin
  result := lightBm.ScanLine[y];
end;
//*********************************************************
function StartLissaDance: string;
begin
  result := 'lissa dance';
//  SetGraphMode (pf15bit);
  SetGraphMode (pf16bit);
//  SetGraphMode (pf24bit);
//  SetGraphMode (pfDevice);
//  SetGraphMode (pf8bit);
  ResizeLightningMap;
end;
//*********************************************************
// calculate simple diamont lightning map
//---------------------------------------------------------
procedure CalculateLightBitmap1;
var   xi,yi, ly,li: longint;
      lp: ^byte;
begin
  for yi := 0 to pageY*2-1          //        ______maxLI
  do begin                          //       ^
    ly := maxLI - abs(pageY - yi);  //     /   \
    lp := lightBm.ScanLine[yi];     // __/       \__minLI
    for xi := 0 to pageX*2-1
    do begin
      li := (ly + maxLI - abs(pageX - xi)) div 2;
      if li <= minLI
      then lp^ := minLI
      else lp^ := li;
      inc (lp);
    end;
  end;
end;
//*********************************************************
// calculate simple circle lightning map
//---------------------------------------------------------
procedure CalculateLightBitmap2;
var   xi,yi, ly,li: longint;
      lp: ^byte;
begin
  for yi := 0 to pageY*2-1          //        ______maxLI
  do begin                          //       ^
    ly := 2 * sqr(yi - pageY);      //     /   \
    lp := lightBm.ScanLine[yi];     // __/       \__minLI

    for xi := 0 to pageX*2-1
    do begin
      li := maxLI - trunc(sqrt(ly + 2*sqr(xi-pageX)));
      if li <= minLI
      then lp^ := minLI
      else lp^ := li;
      inc (lp);
    end;
  end;
end;
//*********************************************************
// create / resize lightning map
//---------------------------------------------------------
procedure ResizeLightningMap;
begin
  if lightBm = nil
  then begin
    lightBm := tBitmap.Create;
    lightBm.PixelFormat := pf8Bit;
  end;
  // resize lightning map to double window size
  lightBm.Width  := 2*pageX;
  lightBm.Height := 2*pageY;
  SetIntensityPalette (lightBm);
//  CalculateLightBitmap1;
  CalculateLightBitmap2;
end;

//*********************************************************
// animate precalculated intensity light map
// input:      dm: destination bitmap
//---------------------------------------------------------
procedure Dancing (var dm: tBitmap);
var
    lx1,ly1,                 // light position
    lx2,ly2,
    lx3,ly3,
    xi,yi: longint;          // pixel position
    rgb3Ptr: ^byte;          // drawing pointer
    rgb2Ptr: ^word;          // rgb15 or rgb16 coded
    lp1,lp2,lp3: ^byte;      // light intensity pointer
begin
  // calculate light position
  lx1 := pageX div 2 + round(cos(1.1*picNo*moveSpeed / 256) * radius);
  ly1 := pageY div 2 + round(sin(1.2*picNo*moveSpeed / 200) * radius);
  lx2 := pageX div 2 + round(cos(1.3*picNo*moveSpeed / 256) * radius);
  ly2 := pageY div 2 + round(sin(1.4*picNo*moveSpeed / 200) * radius);
  lx3 := pageX div 2 + round(cos(1.5*picNo*moveSpeed / 256) * radius);
  ly3 := pageY div 2 + round(sin(1.6*picNo*moveSpeed / 200) * radius);

  for yi := 0 to GetMaxY
  do begin
    rgb3Ptr  := dm.ScanLine[yi];
    rgb2Ptr := pointer(rgb3Ptr);
{$R-,Q-}
    lp1 := lightBm.ScanLine[yi + pageY - ly1];   inc (lp1, pageX-lx1);
    lp2 := lightBm.ScanLine[yi + pageY - ly2];   inc (lp2, pageX-lx2);
    lp3 := lightBm.ScanLine[yi + pageY - ly3];   inc (lp3, pageX-lx3);
    case dm.PixelFormat of
    pf24Bit:
      for xi := 0 to GetMaxX
      do begin
        rgb3Ptr^ := lp1^;  inc (rgb3Ptr);   inc(lp1);
        rgb3Ptr^ := lp2^;  inc (rgb3Ptr);   inc(lp2);
        rgb3Ptr^ := lp3^;  inc (rgb3Ptr);   inc(lp3);
      end;

    pf16Bit:
      for xi := 0 to GetMaxX
      do begin
        rgb2Ptr^ := RGB16(lp1^,lp2^,lp3^);
        inc(rgb2Ptr);
        inc(lp1);
        inc(lp2);
        inc(lp3);
      end;

    pf15Bit:
      for xi := 0 to GetMaxX
      do begin
        rgb2Ptr^ := RGB15(lp1^,lp2^,lp3^);
        inc(rgb2Ptr);
        inc(lp1);
        inc(lp2);
        inc(lp3);
      end;
    else begin
      ErrorMessage ('unexpected pixel format');
      exit;
      end;
    end;
  end;
end;

//*********************************************************
// display transparent shadowed help text
//---------------------------------------------------------
procedure DisplayHelp;
var   h: word;
begin
  TextColor (clWhite);
  h := TextHeight;
  ShadowTextXY (10,1*h, ' i/j    moving speed ' + IntToStr(moveSpeed));
  ShadowTextXY (10,2*h, ' o/k  moving radius ' + IntToStr(radius));
end;
//*********************************************************
// draw one picture and handle keyboard input
//---------------------------------------------------------
procedure LissaDance;
begin
  Dancing (bgBitmap);
  if animHelp then DisplayHelp;
  case cb of
    'i': inc(moveSpeed);
    'j': dec(moveSpeed);
    'o': inc(radius,2);
    'k': dec(radius,2);
  end;
  cb := #0;
end;

end.

