{*******************************************************************}
{*                                                                 *}
{*  UNIT:     G r a p h R G B               23.12.97,  31.08.99    *}
{*                                                                 *}
{*                                                                 *}
{*  Extended graphic unit for true color handling                  *}
{*                                                                 *}
{*            Copyright (c) 1999 by Gerd Platl                     *}
{*                                                                 *}
{*  --- functions and procedures ---                               *}
{*                                                                 *}
{*  RGB2Color           convert r g b values to longint color      *}
{*  Color2RGB           convert longint color to r g b values      *}
{*  SetColor            set active color                           *}
{*  SetRGBColor         set active color                           *}
{*  GetRGBColor         get active color                           *}
{*  SetRGBIntensity     set intensity of color drawing             *}
{*  GetRGBIntensity     return actual color intensity              *}
{*                                                                 *}
{*  SetColorPixel       draw pixel with active color and intensity *}
{*  PutColorPixel       draw pixel with given color                *}
{*  GetColorPixel       get pixel color                            *}
{*  PutPixelRGB         draw pixel with given R G B values         *}
{*  GetPixelRGB         get R G B values of pixel                  *}
{*                                                                 *}
{*  Blend...            Blending                                   *}
{*  AA...               Anti-Alaising                              *}
{*  MipMap              generate Multiple Resolution Picture       *}
{*                                                                 *}
{*  GouraudHLine        Zeichne horizontal Linie mit Farbverlauf   *}
{*  GouraudRectangle    Zeichne Rechteck mit Farbverlauf           *}
{*  DrawRGBColors       draw RGB colors in rectangle               *}
{*                                                                 *}
{*******************************************************************}
{$I tpc_opt.inc}
{$X+}

{$ifdef test}
  {$define xtest1   main function test}
  {$define xtest2   sub function test}
  {$define xtest3   color function test}
{$endif}

unit GraphRGB;

INTERFACE  uses ColorRGB;

{32 bit color}
type  tRGBIColor = packed record
        b,g,r, i: byte;
      end;

type  tColor = longint;

{windows colors:        i b g r}

const clBlack        = $00000000;
      clBlue         = $00FF0000;
      clGreen        = $0000FF00;
      clCyan         = $00FFFF00;
      clRed          = $000000FF;
      clMagenta      = $00FF00FF;
      clYellow       = $0000FFFF;
      clWhite        = $00FFFFFF;

{intensity colors:      i r g b}
const iBlack        = $FF000000;
      iBlue         = $FF00007F;
      iGreen        = $FF007F00;
      iCyan         = $FF007F7F;
      iRed          = $FF7F0000;
      iMagenta      = $FF7F007F;
      iYellow       = $FFFFFF00;
      iWhite        = $FFFFFFFF;

      iDarkGray     = $FF7F7F7F;
      iLightGray    = $FFCFCFCF;
      iLightBlue    = $FF7F7FFF;
      iLightGreen   = $FF3FFF3F;
      iLightCyan    = $FF3FFFFF;
      iLightRed     = $FFFF3F3F;
      iLightMagenta = $FFFF3FFF;
      iBrown        = $FF444400;

{-- color handling --}
function  RGB2Color (r,g,b: byte): longint;
procedure Color2RGB (color: tColor; var r,g,b: byte);
procedure SetColor (color: tColor);
function GetColor: longint;
procedure SetRGBColor (r,g,b: byte);
procedure GetRGBColor (var r,g,b: byte);
procedure SetRGBIntensity (i: byte);
function  GetRGBIntensity: byte;

{-- standard routines --}
procedure SetColorPixel (x,y: integer);
procedure PutColorPixel (x,y: integer; col: tColor);
function  GetColorPixel (x,y: integer): tColor;
procedure GetPixelRGB (x,y: integer; var r,g,b: byte);
procedure PutPixelRGB (x,y: integer; r,g,b: byte);
procedure HLine (x1,x2,y: integer);
procedure VLine (x,y1,y2: integer);
procedure Line (x1,y1, x2,y2: integer);

{-- blending --}
procedure BlendPixel (x,y: integer; int: byte);
procedure BlendHLine (x1,x2,y: integer; int: byte);
procedure BlendVLine (x,y1,y2: integer; int: byte);
procedure BlendRectangle (x1,y1, x2,y2: integer; int: byte);
procedure BlendBar (x1,y1, x2,y2: integer; int: byte);

{-- anti-aliasing --}
procedure AAHLine (x1,x2,y: integer);
procedure AAVLine (x,y1,y2: integer);
procedure AALine (x1,y1, x2,y2: integer);
procedure AARectangle (x1, y1, x2, y2: integer);
procedure AACircle (mx,my, radius: integer);
procedure AAEllipse (mx,my, a,b: integer);
procedure AADisk (mx,my, a,b: integer);
procedure MipMap (x,y,xl,yl: word);

{-- Gouraud Shading --}
procedure GouraudHLine (x1,x2,y: integer; c1,c2: tRGB);
procedure GouraudRectangle (x1,y1,x2,y2: integer;
                            c1,c2,c3,c4: tRGB);
procedure DrawRGBColors (x1,y1,xdim,ydim: integer);


{*********************************************************}
IMPLEMENTATION   uses System2, Crt2, TpString, Graph2;

const _activeColor: tRGBIColor = (b:255; g:255; r:255; i:255);

      GetX: integer = 0;
      GetY: integer = 0;

{*********************************************************}
{ convert r g b values to longint color                   }
{ input:     r,g,b:    rgb values of pixel color (0..255) }
{ return:    longint:  mode specific color value          }
{ example: 15bit (555) 00000000 00000000 0rrrrrgg gggbbbbb}
{          16bit (565) 00000000 00000000 rrrrrggg gggbbbbb}
{          24bit (888) 00000000 rrrrrrrr gggggggg bbbbbbbb}
{---------------------------------------------------------}
function RGB2Color (r,g,b: byte): longint;
begin
  RGB2Color := longint(r) shl 16 + longint(g) shl 8 + b;
end;
{*********************************************************}
{ convert longint color to r g b values                   }
{ input:     color: mode specific color value             }
{ output:    r,g,b: rgb values of pixel color (0..255)    }
{---------------------------------------------------------}
procedure Color2RGB (color: tColor; var r,g,b: byte);
begin
  r := (color shr 16) and $FF;
  g := (color shr  8) and $FF;
  b := (color       ) and $FF;
{
  writeln (tst, 'Color2RGB:  ',r:5,g:5,b:5, IntToHex(color,8):10);
}
end;
{*********************************************************}
procedure SetColor (color: tColor);

begin
  _activeColor := tRGBIColor(color);
{
  writeln (tst, 'SetColor: ',IntToHex(color,8));
}
end;
{*********************************************************}
function GetColor: longint;

begin
  GetColor := longint(_activeColor);
end;
{*********************************************************}
{ input:     r,g,b: rgb values of pixel color (0..255)    }
{---------------------------------------------------------}
procedure SetRGBColor (r,g,b: byte);
begin
  _activeColor.r := r;
  _activeColor.g := g;
  _activeColor.b := b;
{
  writeln (tst, 'SetRGBColor:',r:4,g:5,b:5
              , IntToHex(longint(_activeColor),8):10);
}
end;
{*********************************************************}
{ output:    r,g,b: rgb values of active color (0..255)   }
{---------------------------------------------------------}
procedure GetRGBColor (var r,g,b: byte);
begin
  r := _activeColor.r;
  g := _activeColor.g;
  b := _activeColor.b;
end;
{*********************************************************}
{ set intensity of active color.
{ input:     i: byte;  color intensity (0..255)
{---------------------------------------------------------}
procedure SetRGBIntensity (i: byte);
begin
  _activeColor.i := i;
end;
{*********************************************************}
{ get the actual color intensity
{ return     i: byte;  actual color intensity (0..255)
{---------------------------------------------------------}
function GetRGBIntensity: byte;
begin
  GetRGBIntensity := _activeColor.i;
end;

{*********************************************************}
{ get rgb values of pixel at position x,y                 }
{ if position is out of range, red will be returned       }
{ input:     x,y:   pixel position                        }
{ output:    r,g,b: rgb values of pixel color             }
{---------------------------------------------------------}
procedure GetPixelRGB (x,y: integer; var r,g,b: byte);
var   color: longint;
begin
  if  (x >= 0) and (x <= GetMaxX)
  and (y >= 0) and (y <= GetMaxY)
  then _GetPixelRGB (x,y, r,g,b)
  else begin
    r := 255; g := 0; b := 0;
  end;
{
  writeln (tst, 'GetPixelRGB: ',r:5,g:5,b:5, IntToHex(color,8):10);
}
end;
{*********************************************************}
{ draw pixel in active viewport                           }
{ input:     x,y:   pixel position                        }
{            r,g,b: rgb values of pixel color             }
{---------------------------------------------------------}
procedure PutPixelRGB (x,y: integer; r,g,b: byte);

begin
  if x < 0 then exit;
  if y < 0 then exit;
  inc (x, _wl);  if x > _wr then exit; { ===> }
  inc (y, _wo);  if y > _wu then exit; { ===> }
  _PutPixelRGB (x, y, r,g,b);
end;
{*********************************************************}
{ input:     x,y:   pixel position                        }
{ output:    color: true color value of pixel             }
{---------------------------------------------------------}
function GetColorPixel (x,y: integer): longint;
var   r,g,b: byte;
begin
  if  (x >= 0) and (x <= GetMaxX)
  and (y >= 0) and (y <= GetMaxY)
  then begin
    _GetPixelRGB (x,y, r,g,b);
    GetColorPixel := RGB2Color(r,g,b);
  end
  else GetColorPixel := $00FF0000;   {red}
{
  writeln (tst, 'GetColorPixel(',x,',',y,')=$', IntToHex(color,8):10);
}
end;
{*********************************************************}
{ draw pixel with active color and intensity              }
{ input:     x,y:   pixel position                        }
{            col:   true color value                      }
{---------------------------------------------------------}
procedure SetColorPixel (x,y: integer);
var   r,g,b: byte;
begin
  if x < 0 then exit;
  if y < 0 then exit;
  inc (x, _wl);  if x > _wr then exit; { ===> }
  inc (y, _wo);  if y > _wu then exit; { ===> }
{
  writeln (tst, 'SetPixelRGB :', x:5, y:5, r:5, g:5, b:5);
}
  with _activeColor
  do _PutPixelRGB (x, y, r,g,b);
end;
{*********************************************************}
{ input:     x,y:   pixel position                        }
{            col:   true color value                      }
{---------------------------------------------------------}
procedure PutColorPixel (x,y: integer; col: tColor);
var   r,g,b: byte;
begin
  if x < 0 then exit;
  if y < 0 then exit;
  Color2RGB (col, r,g,b);
{
  writeln (tst, 'PutPixelRGB :', x:5, y:5, r:5, g:5, b:5);
}
  _PutPixelRGB (x, y, r,g,b);
end;
{*********************************************************}
{ mix pixel at position x,y with the active drawing color }
{ input:     x,y:   pixel position                        }
{            int:   intensity (0..255)                    }
{ global:    _activeColor: active rgb color               }
{ formula:   c := (c_old * (255-i) + c_new * i) / 255     }
{---------------------------------------------------------}
procedure BlendPixel (x,y: integer; int: byte);
var   r2,g2,b2: byte; i1,i2: longint;   r3,g3,b3: longint;
begin
  if int = 255
  then begin
    PutColorPixel (x,y, longint(_activeColor));
    exit;
  end
  else if int = 0 then exit;

  i1 := int;
  i2 := 255 - int;
  GetPixelRGB (x,y, r2,g2,b2);

  with _activeColor
  do begin
   r3 := (255 + r * i1 + r2 * i2) shr 8;
   g3 := (255 + g * i1 + g2 * i2) shr 8;
   b3 := (255 + b * i1 + b2 * i2) shr 8;
{
  writeln (tst, 'BlendPixel: ', r:5, g:5, b:5,' i=',int:3,#13#10
              , '            ',r2:5,g2:5,b2:5,#13#10
              , '            ',r3:5,g3:5,b3:5);
}
  end;
  PutPixelRGB (x,y, r3,g3,b3);
{
  PutPixelRGB (x,y, (255 + r * i1 + r2 * i2) shr 8
                  , (255 + g * i1 + g2 * i2) shr 8
                  , (255 + b * i1 + b2 * i2) shr 8);
}
end;

{*********************************************************}
{ draw 4 edge points of a rectangle using alpha blend
ing  }
{ input:     x1,y1:    left top point                     }
{            x2,y2:    right button point                 }
{            i:        intensity (0..255)                 }
{---------------------------------------------------------}
procedure QuadroPixel (x1,y1,x2,y2: integer; int: byte);
begin
{
  writeln (tst, 'QuadroPixel: ', x1:5, y1:5, x2:5,y2:5,' int=',int:3);
}
  with _activeColor
  do begin
    BlendPixel (x1,y1, int);
    BlendPixel (x2,y1, int);
    BlendPixel (x1,y2, int);
    BlendPixel (x2,y2, int);
  end;
end;
{*********************************************************}
{ draw horizontal line with clipping                      }
{---------------------------------------------------------}
procedure HLine (x1,x2,y: integer);
var   xi: integer;
begin
  if x1 > x2 then SwapInt(x1,x2);
{
  if x1 < actViewPort.x1 then x1 := actViewPort.x1;
  if x2 > actViewPort.x2 then x2 := actViewPort.x2;
}
  for xi := x1 to x2
  do PutColorPixel (xi,y, longint(_activeColor));
end;
{*********************************************************}
{ draw vertical line with clipping                        }
{---------------------------------------------------------}
procedure VLine (x,y1,y2: integer);
var   yi: integer;
begin
  if y1 > y2 then SwapInt(y1,y2);
{
  if x1 < actViewPort.x1 then x1 := actViewPort.x1;
  if x2 > actViewPort.x2 then x2 := actViewPort.x2;
}
  for yi := y1 to y2
  do PutColorPixel (x,yi, longint(_activeColor));
end;
{*********************************************************}
procedure BlendHLine (x1,x2,y: integer; int: byte);
var   xi: integer;
begin
  if int = 255 then HLine (x1,x2,y)
  else if int = 0 then exit;
  if x1 > x2 then SwapInt(x1,x2);
  for xi := x1 to x2 do BlendPixel (xi,y, int);
end;
{---------------------------------------------------------}
procedure BlendVLine (x,y1,y2: integer; int: byte);
var   yi: integer;
begin
  if      int = 255 then VLine (x,y1,y2)
  else if int =   0 then exit;         { ===> }
  if y1 > y2 then SwapInt(y1,y2);
  for yi := y1 to y2 do BlendPixel (x,yi, int);
end;
{*********************************************************}
{ draw anti-aliased horizontal line                       }
{---------------------------------------------------------}
procedure AAHLine (x1,x2,y: integer);
begin
  if x1 > x2 then system2.SwapInt (x1,x2);
  BlendHLine (x1,x2,y-1, 127);         { 50%}
       HLine (x1,x2,y       );         {100%}
  BlendHLine (x1,x2,y+1, 127);         { 50%}
end;
{*********************************************************}
{ draw anti-aliased vertical line                          }
{---------------------------------------------------------}
procedure AAVLine (x,y1,y2: integer);
begin
  if y1 > y2 then system2.SwapInt (y1,y2);
  BlendVLine (x-1,y1,y2, 127);       { 50%}
       VLine (x,  y1,y2     );       {100%}
  BlendVLine (x+1,y1,y2, 127);       { 50%}
end;
{*********************************************************}
{ draw transparent rectangle                              }
{ input:     x1,y1:    left top edge                      }
{            x2,y2:    right button edge                  }
{            int:      intensity (0..255)                 }
{---------------------------------------------------------}
procedure BlendRectangle (x1,y1, x2,y2: integer; int: byte);
var   si: byte;
begin
{
  writeln (tst, 'BlendRectangle: ', x1:5, y1:5, x2:5, y2:5);
}
  if int = 0 then exit;
  si := _activeColor.i;
  _activeColor.i := int;
  BlendHLine (x1, pred(x2), y1, int);     { ----| }
  BlendVLine (x2, y1, pred(y2), int);     { |   | }
  BlendHLine (succ(x1), x2, y2, int);     { |   | }
  BlendVLine (x1, succ(y1), y2, int);     { |---- }
  _activeColor.i := si;
end;
{*********************************************************}
{ draw rectangle with actual color and 50% blend color    }
{ input:     x1,y1:    left top edge                      }
{            x2,y2:    right button edge                  }
{---------------------------------------------------------}
procedure AARectangle (x1,y1, x2,y2: integer);
var   i: byte;
begin
{
  writeln (tst, 'AARectangle: ', x1:5, y1:5, x2:5, y2:5);
}
  i := _activeColor.i;
  BlendRectangle (x1,   y1,   x2,   y2,   i);       {100%}
  BlendRectangle (x1-1, y1-1, x2+1, y2+1, i shr 1);  {50%}
  BlendRectangle (x1+1, y1+1, x2-1, y2-1, i shr 1);  {50%}
end;
{*********************************************************}
{ draw transparent filled rectangle with actual color     }
{---------------------------------------------------------}
procedure BlendBar (x1, y1, x2, y2: integer; int: byte);

var   xi,yi: integer;
begin
{
  writeln (tst, 'BlendBar: ', x1:5, y1:5, x2:5, y2:5,'  int=',int);
}
  if int = 255
{  then graph.Bar (x1,y2, x2,y2)}
  then for yi := y1 to y2 do      HLine (x1,x2,yi)
  else for yi := y1 to y2 do BlendHLine (x1,x2,yi, int);
end;
{*********************************************************}
{ draw line using Bresenham-algorithm                     }
{ usable for all directions                 MC 11/86 S.12 }
{---------------------------------------------------------}
procedure Line (x1,y1, x2,y2: integer);

var  px,py,pz, dx,dy,dz, i1,i2: longint;

begin
  if y1 > y2
  then begin
    SwapInt (x1,x2);
    SwapInt (y1,y2);
  end;

  if y1 = y2
  then begin
    HLine (x1,x2,y1);                  {horizontal line}
    exit;                              { ===> }
  end;

  if x1 = x2
  then begin
    VLine (x1,y1,y2);                  {vertical line}
    exit;                              { ===> }
  end;

  dx := abs (x2 - x1);
  dy := abs (y2 - y1);

  px := x1;
  py := y1;
  if x1 < x2 then pz := 1 else pz := -1;

  if dx > dy then i2 := dx else i2 := dy;
  dz := i2 div 2;
  SetColorPixel (px,py);
  for i1 := 1 to i2          {y-Schleife}
  do begin
    if dz < dy
    then begin
      dz := dz + dx;
      inc (py);              { y + 1 }
    end;
    if dz >= dy
    then begin
      px := px + pz;         { x  1 }
      dz := dz - dy;
    end;
    SetColorPixel (px,py);
  end;
end;
{*********************************************************}
{ draw anti-aliased line using Bresenham-Algorithm        }
{ with a visual thickness of 2 pixels (looks good !)      }
{ input:     x1,y1:   endpoint 1                          }
{            x2,y2:   endpoint 2                          }
{---------------------------------------------------------}
procedure AALine (x1,y1, x2,y2: integer);

var  px,py,pz, dx,dy,dz, i1: longint;
     c1,c3: longint;

begin
{
  writeln (tst, 'line: ',x1:4, y1:4, x2:4, y2:4);
}
  c1 := longint(_activeColor);                   {100%}
  SetColor (c1);

  if x1 > x2
  then begin
    SwapInt(x1,x2);
    SwapInt(y1,y2);
  end;

  if y1 = y2
  then begin
    AAHLine (x1,x2-1,y1);         {horizontal line}
    exit;                         { ===> }
  end;

  if x1 = x2
  then if (x1 >= 0) and (x1 <= GetMaxX)
  then begin
    if y1 > y2 then SwapInt (y1,y2);
    AAVLine (x1,y1,y2-1);         {vertical line}
    exit;                         { ===> }
  end;

  dx := abs (x2 - x1);
  dy := abs (y2 - y1);
  px := x1;                       {p is startpoint}
  py := y1;
  if y1 < y2 then pz := 1 else pz := -1;
  if dx > dy
  then with _activeColor
  do begin                        {hor. line}
    dec(dx);
    dz := dx div 2;
    PutColorPixel (px,py, c1);    {100%}
    BlendPixel (px,  py-1,  127); { 50%}
    BlendPixel (px,  py+1,  127); { 50%}
    for i1 := 1 to dx             {x-loop}
    do begin
      if dz < dx
      then begin
        dz := dz + dy;
        inc (px);                 { x + 1 }
      end;
      if dz >= dx
      then begin
        py := py + pz;            { y  1 }
        dz := dz - dx;
      end;
      PutColorPixel (px,py, c1);       {100%}
      c3 := byte ((255 * dz) div dx);  {intensity}
      BlendPixel (px,py-pz, 255-c3);   {100%-int}
      BlendPixel (px,py+pz, c3);       {int%}
{
  writeln (tst, px:4, py:4, dx:4, dy:4, dz:4, c3:4, 255-integer(c3):4);
}
    end;
  end
  else with _activeColor
  do begin                        {vert. line}
    dec(dy);
    dz := dy div 2;
    PutColorPixel (px,py, c1);    {100%}
    BlendPixel (px-1,py,   127);  { 50%}
    BlendPixel (px+1,py,   127);  { 50%}
    for i1 := 1 to dy             {y-loop}
    do begin
      if dz < dx
      then begin
        dz := dz + dy;
        inc (px);                 { x + 1 }
      end;
      if dz >= dx
      then begin
        py := py + pz;            { y  1 }
        dz := dz - dx;
      end;
      PutColorPixel (px,py, c1);       {100%}
      c3 := byte ((255 * dz) div dy);  {intensity}
      BlendPixel (px+1,py, 255-c3);    {100%-int}
      BlendPixel (px-1,py, c3);        {int%}
{
  writeln (tst, px:4, py:4, dx:4, dy:4, dz:4, c3:4);
}
    end;
  end;
end;
{*********************************************************}
{ draw line using Bresenham-Algorithm for all directions  }
{ with a visual thickness of 1 pixels (not really nive)   }
{ input:     x1,y1:   endpoint 1                          }
{            x2,y2:   endpoint 2                          }
{---------------------------------------------------------}
procedure AALine1 (x1,y1, x2,y2: integer);

var  px,py,pz, dx,dy,dz, i1: longint;
     c1,c3: longint;

begin
  c1 := longint(_activeColor);                   {100%}
  SetColor (c1);
  if y1 = y2
  then begin
    HLine (x1,x2,y1);           {horizontal line}
    exit;                         { ===> }
  end;

  if x1 = x2
  then begin
    VLine (x1,y1,y2);           {vertical line}
    exit;                         { ===> }
  end;

  dx := abs (x2 - x1);
  dy := abs (y2 - y1);
  if x1 < x2
  then begin
    px := x1;
    py := y1;
    if y1 < y2 then pz := 1 else pz := -1;
  end
  else begin
    px := x2;
    py := y2;
    if y1 > y2 then pz := 1 else pz := -1;
  end;
  if dx > dy
  then with _activeColor
  do begin                        {hor. line}
    dz := dx div 2;
    BlendPixel (px,  py-1,  63);
    BlendPixel (px,  py  ,  63);
    for i1 := 1 to dx             {x-loop}
    do begin
      if dz < dx
      then begin
        dz := dz + dy;
        inc (px);                 { x + 1 }
      end;
      if dz >= dx
      then begin
        py := py + pz;            { y  1 }
        dz := dz - dx;
      end;
      c3 := byte ((255 * dz) div dx);        {intensity}
      BlendPixel (px,py-pz, 255-integer(c3));  {100%-int}
      BlendPixel (px,py, c3);      {int%}
{
  writeln (tst, px:4, py:4, dx:4, dy:4, dz:4, c3:4, 255-integer(c3):4);
}
    end;
  end
  else with _activeColor
  do begin                        {vert. line}
    dz := dy div 2;
    BlendPixel (px-1,py,    63);
    BlendPixel (px  ,py,    63);
    for i1 := 1 to dy             {y-loop}
    do begin
      if dz < dx
      then begin
        dz := dz + dy;
        inc (px);                 { x + 1 }
      end;
      if dz >= dx
      then begin
        py := py + pz;            { y  1 }
        dz := dz - dx;
      end;
      c3 := byte ((255 * dz) div dy);       {intensity}
      BlendPixel (px,py, 255-integer(c3)); {100%-int}
      BlendPixel (px-1,py, c3);     {int%}
{
  writeln (tst, px:4, py:4, dx:4, dy:4, dz:4, c3:4);
}
    end;
  end;
end;
{*********************************************************}
procedure MoveTo (x, y: integer);

begin
  GetX := x;
  GetY := y;
end;
{---------------------------------------------------------}
procedure MoveRel (dx, dy: integer);

begin
  inc (GetX, dx);
  inc (GetY, dy);
end;
{*********************************************************}
procedure AALineTo (x, y: integer);

begin
  AALine (GetX, GetY, x,y);
  GetX := x;
  GetY := y;
end;
{---------------------------------------------------------}
procedure AALineRel (dx, dy: integer);

begin
  AALine (GetX, GetY, GetX + dx, GetY + dy);
  GetX := GetX + dx;
  GetY := GetY + dy;
end;
{*********************************************************}
{ input:     pn:    number of points                      }
{            polyData:   array of pointType               }
{---------------------------------------------------------}
procedure DrawAAPoly (pn: word; var polydata);

var   ni: word;
      po: ^pointType;
      p1: pointType;

begin
  po := @(polyData);
  p1 := po^;                 {Punkt 1 merken}
  MoveTo (p1.x, p1.y);
  for ni := 1 to pn - 1
  do begin
    inc (longint (po), 4);
    AALineTo (po^.x, po^.y);
  end;
  {letzter Punkt ungleich ?}
  if longint (p1) <> longint (po^)
  then AALineTo (p1.x, p1.y);
end;
{*********************************************************}
{ draw anti-aliased ellipse using Bresenham-Algorithm     }
{ with actual color                                       }
{ input:     mx,my:    middle point                       }
{            a,b:      axis length (must be < 2000)       }
{ global:    _activeColor: active drawing color           }
{---------------------------------------------------------}
procedure DrawAAEllipse (mx,my, a,b: integer);

var   x,   x1,x2,  y1,y2: integer;
      aq,bq, dx,dy, r, rx,ry: longint;
      inColor, outColor: integer;      {0..255}
      aay: boolean;
      c1: longint;
begin
  c1 := longint(_activeColor);
  if a = 0
  then if b = 0
  then PutColorPixel (mx, my, c1) {draw just a point}
  else AAVLine (mx, my-b,my+b)    {vertical line}
  else if b = 0
  then AAHLine (mx-a,mx+a, my)    {horizontal line}
  else begin
    x1 := mx - a;   y1 := my;     {init pixel positions}
    x2 := mx + a;   y2 := my;     { for QuadroPixel}

    aq := longint (a) * a;        {calc sqr}
    bq := longint (b) * b;
    dx := aq shl 1;               {dx := 2 * a * a}
    dy := bq shl 1;               {dy := 2 * b * b}
    r  := a * bq;                 {r  := a * b * b}
    rx := r shl 1;                {rx := 2 * a * b * b}
    ry := 0;                      {because y = 0}
    x  := a;
    aay:= false;                  {set aa-points left&right}

    PutColorPixel (x1, my, c1);   {draw highest point}
    PutColorPixel (x2, my, c1);   {draw lowest point}
    BlendPixel (x1-1, my, 127);   {50%}
    BlendPixel (x1+1, my, 127);
    BlendPixel (x2-1, my, 127);
    BlendPixel (x2+1, my, 127);

    while x > 1
    do begin
{
  PutColorPixel (x1, getmaxY div 2+r div 100000, c1);
  writeln (tst, ' r=',r:10,' rx=',rx:10, ' ry=',ry:10, ' rx-r=',rx-r:10);
}
      if r > 0
      then begin                  { y + 1 }
        inc (y1);   dec (y2);
        inc (ry, dx);             {ry = dx * y}
        dec (r, ry);              {r = r - dx * y}
      end;
      if r <= 0
      then begin                  { x - 1 }
        dec (x);
        inc (x1);   dec (x2);
        dec (rx, dy);             {rx = dy * x}
        inc (r, rx);              {r = r + dy * x}
      end;
      QuadroPixel (x1,y1,x2,y2, 255);  {draw mirror points}

      {-- anti-aliasing pixel drawing --}
      {inColor  := 255 * (rx-r) div (rx+ry);
       because of integer overflow for greater ellipses
       this statement must use floating point operations
      }
      inColor  := trunc(255.0 * (rx-r) / (rx+ry));   {!!! very tricky !!!}
      outColor := 255 - inColor;        {outside color intensity}

      if aay
      then begin                  {draw above, below}
        QuadroPixel (x1, y1-1, x2, y2+1, inColor);
        QuadroPixel (x1, y1+1, x2, y2-1, outColor);
      end
      else begin                  {draw left,right}
        QuadroPixel (x1+1, y1, x2-1, y2, inColor);
        QuadroPixel (x1-1, y1, x2+1, y2, outColor);
        if rx < ry                {tangent -45 reached ?}
        then begin
          aay := true;            {draw diagonal}
          QuadroPixel (x1, y1+1, x2, y2-1, outColor);
        end;
      end;
    end;
    PutColorPixel (mx, y1, c1);   {draw highest point}
    PutColorPixel (mx, y2, c1);   {draw lowest point}
    BlendPixel (mx, y1-1, 127);
    BlendPixel (mx, y1+1, 127);
    BlendPixel (mx, y2-1, 127);
    BlendPixel (mx, y2+1, 127);
  end;
end;
{*********************************************************}
{ draw anti-aliased ellipse disk using Bresenham-Algorithm}
{ input:     mx,my:    middle point                       }
{            a,b:      axis length (must be < 2000)       }
{---------------------------------------------------------}
procedure AADisk (mx,my, a,b: integer);

var   x,   x1,x2,  y1,y2: integer;
      aq,bq, dx,dy, r, rx,ry: longint;
      outColor: byte;
      aay: boolean;

begin
  with _activeColor do
  if a = 0
  then if b = 0
  then PutColorPixel (mx, my, longint(_activeColor))
  else begin
    AAVLine (mx, my-b, my+b-1);
    exit;
  end;                            { ===> }
  if b = 0
  then begin
    AAHLine (mx-a, mx+a-1, my);
    exit;                         { ===> }
  end
  else begin
    x1 := mx - a;   y1 := my;     {init pixel positions}
    x2 := mx + a;   y2 := my;
    HLine (x1, x2, my);
    BlendPixel (x1-1, my, 127);
    BlendPixel (x2+1, my, 127);

    aq := longint (a) * a;        {calc sqr}
    bq := longint (b) * b;
    dx := aq shl 1;               {dx := 2 * a * a}
    dy := bq shl 1;               {dy := 2 * b * b}
    r  := a * bq;                 {r  := a * b * b}
    rx := r shl 1;                {rx := 2 * a * b * b}
    ry := 0;                      {because y = 0}
    x := a;
    aay := false;                 {set aa-points left&right}

    while x > 1
    do begin
      if r > 0
      then begin                  { y + 1 }
        inc (y1);   dec(y2);
        inc (ry, dx);
        dec (r, ry);
        HLine (x1, x2, y1);       {draw 2 horizontal lines}
        HLine (x1, x2, y2);
      end;
      if r <= 0
      then begin                  { x - 1 }
        dec (x);
        inc (x1); dec (x2);
        dec (rx, dy);
        inc (r, rx);
      end;

      {-- Anti-Alaising-Pixel drawing --}
      {outColor := (255 * (r-rx)) div (rx+ry);
       because of integer overflow for greater ellipses
       this statement must use floating point operations}

      outColor := 255-trunc(255.0 * (rx-r) / (rx+ry));
      if aay
      then begin                  {draw above, below}
        QuadroPixel (x1, y1+1, x2, y2-1, outColor);
      end
      else begin                  {draw left,right}
        QuadroPixel (x1-1, y1, x2+1, y2, outColor);
        if rx < ry                {tangent -45 reached ?}
        then begin
          aay := true;            {draw diagonal}
          QuadroPixel (x1, y1+1, x2, y2-1, outColor);
        end;
      end;
    end;
    BlendPixel (mx, y1+1, 127);
    BlendPixel (mx, y2-1, 127);
  end;
end;
{*********************************************************}
{ draw anti-aliased circle line with active color         }
{ input:     mx,my:    middle point                       }
{            radius:   circle radius                      }
{---------------------------------------------------------}
procedure DrawLineEllipse (mx,my, a,b: integer);
var   ni: word;   x1,y1,x2,y2: integer;
{const deg = Pi / 180.0;      1 / Radiant = 0.017453292...}
begin
  x1 := mx+a;
  y1 := my;
  for ni := 1 to 360
  do begin
    x2 := mx + round (cos (ni * pi/180) * a);
    y2 := my + round (sin (ni * pi/180) * b);
    AALine (x1,y1,x2,y2);
    x1 := x2;
    y1 := y2;
  end;
end;
{*********************************************************}
{ draw anti-aliased circle line with active color         }
{ input:     mx,my:    middle point                       }
{            radius:   circle radius                      }
{---------------------------------------------------------}
procedure AACircle (mx,my, radius: integer);
var   xAsp,yAsp: word;  a: integer;
begin
  GetAspectRatio (xAsp,yAsp);
  a := round((radius*yAsp)/xAsp);
  if (a < 2000) and (radius < 2000)
  then DrawAAEllipse (mx,my,a,radius)
  else DrawLineEllipse (mx,my,a,radius);
end;
{*********************************************************}
{ draw anti-aliased ellipse line with active color        }
{ input:     mx,my:    middle point                       }
{            radius:   circle radius                      }
{---------------------------------------------------------}
procedure AAEllipse (mx,my, a,b: integer);
begin
  if (a < 2000) and (b < 2000)
  then DrawAAEllipse (mx,my,a,b)
  else DrawLineEllipse (mx,my,a,b);
end;

{*********************************************************}
{ create from a given image at position x,y of size l*l   }
{ a MIP-Map with half anti-aliased pictures               }
{ input:     x,y:    left top image position              }
{            xl,yl:  image size in pixels                 }
{---------------------------------------------------------}
procedure MipMap (x,y,xl,yl: word);

var   xdim: word;
      p1,p2,p3: tLinePtr;

 {--------------------------------------------------------}
 procedure MipMapX (x,y,xl,yl: word);

 var   l2,x2,y2,xi,yi,xp: word;

       r1,g1,b1: byte;
       r2,g2,b2: byte;

 begin
   l2 := xl shr 1;    {/2}
   inc (xdim, l2);
   x2 := x + xl;
   for yi := 0 to yl-1
   do begin
     y2 := y + yi;
     for xi := 0 to l2-1
     do begin
       xp := x + xi shl 1; {*2}
       GetPixelRGB (xp,   y2, r1,g1,b1);
       GetPixelRGB (xp+1, y2, r2,g2,b2);
       PutPixelRGB (x2+xi,y2, (r1+r2) shr 1, (g1+g2) shr 1, (b1+b2) shr 1);
     end;
   end;
   if l2 > 2 then MipMapX (x2,y,l2,yl);
 end;
 {-------------------------------------------------------}
 procedure MipMapY (x,y,xl,yl: word);

 var   l2,y2,xi,yi,yp: word;

 begin
   l2 := yl shr 1;        {yl/2}
   y2 := y + yl;
   for yi := 0 to l2-1
   do begin
     yp := y + yi shl 1;  {y+yi*2}
     GetLine (p1^, x,yp,   xl);
     GetLine (p2^, x,yp+1, xl);
     for xi := 0 to (xl-1) * bytesPerPixel
     do p3^[xi] := (p1^[xi] + p2^[xi]) div 2;
     PutLine (p3^, x,y2,xl);
     inc (y2);
   end;
   if l2 > 2 then MipMapY (x,y+yl,xl,l2);
 end;
{---------------------------------------------------------}
begin
  GetMem (p1, maxLineBytes);
  GetMem (p2, maxLineBytes);
  GetMem (p3, maxLineBytes);
  xdim := xl;
  MipMapX (x,y,xl,yl);
  MipMapY (x,y,xdim,yl);
  FreeMem (p3, maxLineBytes);
  FreeMem (p2, maxLineBytes);
  FreeMem (p1, maxLineBytes);
end;

{*********************************************************}
{ Mischt die beiden Farben c1 und c2 mit der Intesitt 50%}
{ zusammen und gibt den berechneten Farbwert mix zurck.  }
{   mix := (c1 * 128 + c2 * 128) div 256                  }
{---------------------------------------------------------}
procedure MixHalfColor (var mix, c1,c2: tColor);
const halfMask = $007F7F7F;
begin
  mix := (c1 shr 1) and halfMask
        +(c2 shr 1) and halfMask;
end;
{*********************************************************}
{ Mischt die beiden Farben c1 und c2 mit der Intesitt int}
{ zusammen und gibt den berechneten Farbwert mix zurck.  }
{ Bereich der Intensittsangabe von int: 0..255           }
{   mix := (c1 * (256-int) + c2 * int) div 256            }
{---------------------------------------------------------}
{R-,Q-}
procedure MixColor (var mix, c1,c2: tColor; int: byte);
var  r1,r2: longint;
begin
  r1 := 256 - int;
  r2 := int;
  mix := ((c1 shr 16 and $FF) * r1
        + (c2 shr 16 and $FF) * r2) and $FF00 shl 8
        +((c1 shr  8 and $FF) * r1
        + (c2 shr  8 and $FF) * r2) and $FF00
        +((c1        and $FF) * r1
        + (c2        and $FF) * r2) shr 8;
end;
{*********************************************************}
{ Zeichnet eine horizontale Linie von x1/y nach x2/y, wobei
{ der Farbwert von c1 nach c2 linear interpoliert wird.
{---------------------------------------------------------}
procedure GouraudHLine (x1,x2,y: integer; c1,c2: tRGB);

const mf = 256;
var   r,g,b, dr,dg,db, xi,dx: longint;

begin
  dx := (x2 - x1) * mf;
  if dx <= 0 then exit;   { ===> }
  r := longint(c1.rb) * mf;   dr := (longint(c2.rb) * mf - r) * mf div dx;
  g := longint(c1.gb) * mf;   dg := (longint(c2.gb) * mf - g) * mf div dx;
  b := longint(c1.bb) * mf;   db := (longint(c2.bb) * mf - b) * mf div dx;
  if _activeColor.i = 255
  then for xi := x1 to x2
  do begin
    _PutPixelRGB (xi, y, r shr 8, g div mf, b div mf);
    inc (r,dr);
    inc (g,dg);
    inc (b,db);
  end
  else for xi := x1 to x2
  do begin
    SetRGBColor (r shr 8, g div mf, b div mf);
    BlendPixel (xi,y, 127);
    inc (r,dr);
    inc (g,dg);
    inc (b,db);
  end;
end;
{*********************************************************}
{ Zeichnet ein geflltes Rechteck (x1,y1 .. x2,y2)
{ mit interpolierten RGB-Werten:
{
{  c1+========+c2
{    |        |
{ cy1+..+cx...+cy2
{    |        |
{  c3+========+c4
{
{---------------------------------------------------------}
procedure GouraudRectangle (x1,y1,x2,y2: integer;
                            c1,c2,c3,c4: tRGB);

var  yi,yp, ydim: longint;
     yint: byte;
     cy1,cy2: tRGB;

begin
  ydim := y2 - y1 + 1;
  for yi := 0 to ydim - 1
  do begin
    yp := y1 + yi;
    yint := byte(longint(yi * 255) div ydim);
    MixRGB (cy1, c1,c3, yint);
    MixRGB (cy2, c2,c4, yint);
    GouraudHLine (x1,x2,yp, cy1,cy2);
  end;
end;

{*********************************************************}
{ Zeichnet ein geflltes Rechteck (xdim * ydim)           }
{ mit RGB-Pixel lt. Tabelle an der Position (x1,y1).      }
{                                                         }
{ x1/y1               S c h w a r z                       }
{   +===============================================+     }
{   |       |       |       |       |       |       |     }
{   |       |       |       |       |       |       |     }
{ Magenta--ROT-----Gelb---GRN----Cyan----BLAU--Magenta   }
{   |       |       |       |       |       |       |     }
{   |       |       |       |       |       |       |     }
{   +===============================================+     }
{                       W e i s s                         }
{---------------------------------------------------------}
procedure DrawRGBColors (x1,y1,xdim,ydim: integer);

const colors: array[0..8] of tRGB
      = ((rb:  0; gb:  0; bb:  0)
        ,(rb:255; gb:  0; bb:255)
        ,(rb:255; gb:  0; bb:  0)
        ,(rb:255; gb:255; bb:  0)
        ,(rb:  0; gb:255; bb:  0)
        ,(rb:  0; gb:255; bb:255)
        ,(rb:  0; gb:  0; bb:255)
        ,(rb:255; gb:  0; bb:255)
        ,(rb:255; gb:255; bb:255));

var   xd,yd,x2,y2,ri: word;   c1,c2: tRGB;

begin
  InitCubeColors;
  xd := xdim div 6;
  yd := ydim div 2;
  y2 := y1+ydim-1;
  for ri := 0 to 5
  do begin
    c1 := colors[ri+1];
    c2 := colors[ri+2];
    x2 := x1+ri*xd;
    GouraudRectangle (x2, y1,    x2+xd-1, y1+yd
                     ,colors[0], colors[0], c1,c2);  {black .. rgb}
    GouraudRectangle (x2, y1+yd+1, x2+xd-1, y2
                     ,c1,c2, colors[8], colors[8]);  {rgb .. white}
  end;
end;
{*********************************************************}
end.
